import aiohttp

from typing import Any, Dict
from furl import furl


class BaseAPIRepository:
    _url: furl

    @property
    def url(self):
        return self._url.copy()

    @staticmethod
    async def _post(url: str, headers: Dict[str, str], data: Dict[str, Any]) -> Dict[str, Any]:
        async with aiohttp.ClientSession(headers=headers) as session:
            async with session.post(url, json=data) as r:
                return await r.json()



from typing import List

from defisdk.entities import (
    AdapterBalance, AdapterMetadata, AssetBalance, ProtocolBalance, ProtocolMetadata, TokenBalance, TokenMetadata
)
from defisdk.utils import (
    hash_to_address, hash_to_decimal, hash_to_int, hash_to_list, hash_to_words, words_to_list, words_to_string
)


def defi_sdk_token_metadata_to_entity(words: List[str]) -> TokenMetadata:
    name_location = hash_to_int(words[1]) // 32
    symbol_location = hash_to_int(words[2]) // 32
    return TokenMetadata(
        address=hash_to_address(words[0]),
        name=words_to_string(words[name_location:symbol_location]),
        symbol=words_to_string(words[symbol_location:]),
        decimals=hash_to_int(words[3])
    )


def defi_sdk_token_balance_to_entity(words: List[str]) -> TokenBalance:
    metadata_location = hash_to_int(words[0]) // 32
    metadata = defi_sdk_token_metadata_to_entity(words[metadata_location:])
    return TokenBalance(
        metadata=metadata,
        balance=hash_to_decimal(words[1], metadata.decimals)
    )


def defi_sdk_asset_balance_to_entity(words: List[str]) -> AssetBalance:
    base_location = hash_to_int(words[0]) // 32
    underlying_location = hash_to_int(words[1]) // 32
    base_balance = defi_sdk_token_balance_to_entity(words[base_location:underlying_location])
    underlying_balances = words_to_list(words[underlying_location:], defi_sdk_token_balance_to_entity, True)
    return AssetBalance(
        base_token_balance=base_balance,
        underlying_token_balances=underlying_balances
    )


def defi_sdk_adapter_metadata_to_entity(words: List[str]) -> AdapterMetadata:
    type_location = hash_to_int(words[1]) // 32
    return AdapterMetadata(
        address=hash_to_address(words[0]),
        type=words_to_string(words[type_location:])
    )


def defi_sdk_adapter_balance_to_entity(words: List[str]) -> AdapterBalance:
    metadata_location = hash_to_int(words[0]) // 32
    balances_location = hash_to_int(words[1]) // 32
    metadata = defi_sdk_adapter_metadata_to_entity(words[metadata_location:balances_location])
    balances = words_to_list(words[balances_location:], defi_sdk_asset_balance_to_entity, True)
    return AdapterBalance(
        metadata=metadata,
        asset_balances=balances
    )


def defi_sdk_protocol_metadata_to_entity(words: List[str]) -> ProtocolMetadata:
    name_location = hash_to_int(words[0]) // 32
    description_location = hash_to_int(words[1]) // 32
    website_url_location = hash_to_int(words[2]) // 32
    icon_url_location = hash_to_int(words[3]) // 32
    return ProtocolMetadata(
        name=words_to_string(words[name_location:description_location]),
        description=words_to_string(words[description_location:website_url_location]),
        website_url=words_to_string(words[website_url_location:icon_url_location]),
        icon_url=words_to_string(words[icon_url_location:]),
        version=hash_to_int(words[4])
    )


def defi_sdk_protocol_balance_to_entity(words: List[str]) -> ProtocolBalance:
    metadata_location = hash_to_int(words[0]) // 32
    balances_location = hash_to_int(words[1]) // 32
    metadata = defi_sdk_protocol_metadata_to_entity(words[metadata_location:balances_location])
    balances = words_to_list(words[balances_location:], defi_sdk_adapter_balance_to_entity, True)
    return ProtocolBalance(
        metadata=metadata,
        adapter_balances=balances
    )


def defi_sdk_protocol_balances_to_entity(data: str) -> List[ProtocolBalance]:
    return hash_to_list(data, defi_sdk_protocol_balance_to_entity, dynamic_elements=True)


def defi_sdk_adapter_balances_to_entity(data: str) -> List[AdapterBalance]:
    return hash_to_list(data, defi_sdk_adapter_balance_to_entity, dynamic_elements=True)


def defi_sdk_token_adapter_names_to_list_of_string(data: str) -> List[str]:
    return hash_to_list(data, words_to_string, dynamic_elements=True)


def defi_sdk_protocol_names_to_list_of_string(data: str) -> List[str]:
    return hash_to_list(data, words_to_string, dynamic_elements=True)


def defi_sdk_full_token_balance_to_entity(data: str) -> AssetBalance:
    asset_balance = defi_sdk_asset_balance_to_entity(hash_to_words(data)[1:])
    return asset_balance



from typing import List, Union

from .entities import AdapterBalance, AssetBalance, ProtocolBalance, TokenBalance, TokenMetadata
from .repositories import DeFiSDKAPIRepository
from .settings import DEFI_SDK_REGISTRY


class DeFiSDK:

    def __init__(self, ethereum_node_url: str, defisdk_registry: str = DEFI_SDK_REGISTRY):
        self._repository = DeFiSDKAPIRepository(ethereum_node_url, defisdk_registry=defisdk_registry)

    async def get_account_balance(
            self,
            address: str,
            block: Union[str, int] = 'latest'
    ) -> List[ProtocolBalance]:
        return await self._repository.get_account_balance(address, block)

    async def get_adapter_balance(
            self,
            address: str,
            adapter: str,
            block: Union[str, int] = 'latest'
    ) -> AdapterBalance:
        return await self._repository.get_adapter_balance(address, adapter, block)

    async def get_protocol_balance(
            self,
            address: str,
            protocol_name: str,
            block: Union[str, int] = 'latest'
    ) -> ProtocolBalance:
        return await self._repository.get_protocol_balance(address, protocol_name, block)

    async def get_protocol_names(
            self,
            block: Union[str, int] = 'latest'
    ) -> List[str]:
        return await self._repository.get_protocol_names(block)

    async def get_token_adapter_names(
            self,
            block: Union[str, int] = 'latest'
    ) -> List[str]:
        return await self._repository.get_token_adapter_names(block)

    async def get_protocol_adapter_names(
            self,
            protocol_name: str,
            block: Union[str, int] = 'latest'
    ) -> List[str]:
        return await self._repository.get_protocol_adapter_names(protocol_name, block)

    async def get_token_adapter(
            self,
            token_adapter_name: str,
            block: Union[str, int] = 'latest'
    ) -> str:
        return await self._repository.get_token_adapter(token_adapter_name, block)

    async def get_full_token_balance(
            self,
            token_type: str,
            token_address: str,
            block: Union[str, int] = 'latest'
    ) -> AssetBalance:
        return await self._repository.get_full_token_balance(token_type, token_address, block)

    async def get_token_components(
            self,
            token_type: str,
            token_address: str,
            block: Union[str, int] = 'latest'
    ) -> List[TokenBalance]:
        return await self._repository.get_token_components(token_type, token_address, block)

    async def get_token_metadata(
            self,
            token_type: str,
            token_address: str,
            block: Union[str, int] = 'latest'
    ) -> TokenMetadata:
        return await self._repository.get_token_metadata(token_type, token_address, block)



from binascii import unhexlify
from decimal import Decimal
from typing import Any, Callable, List, Union

from sha3 import keccak_256


def remove_0x_prefix(passed_hash: str) -> str:
    return passed_hash[2:] if passed_hash.startswith("0x") else passed_hash


def represent_address(address: str) -> str:
    address = address.lower()
    return address if address.startswith('0x') else f'0x{address}'


def represent_hash(passed_hash: Union[bytes, str]) -> str:
    passed_hash = passed_hash.hex() if isinstance(passed_hash, bytes) else passed_hash
    return passed_hash if passed_hash.startswith('0x') else f'0x{passed_hash}'


def hash_to_signature(passed_hash: str) -> str:
    passed_hash = remove_0x_prefix(passed_hash)
    return represent_hash(passed_hash[:8])


def hash_to_int(passed_hash: str, unsigned: bool = True) -> int:
    passed_hash = remove_0x_prefix(passed_hash)[:64]
    passed_hash = passed_hash.zfill(64)
    if not unsigned and int(passed_hash[0], 16) >= 8: # negative int
        return int(passed_hash, 16) - 2**256
    return int(passed_hash, 16)


def hash_to_decimal(passed_hash: str, decimals: int = 18) -> Decimal:
    value = hash_to_int(passed_hash)
    sign = 0 if value > 0 else 1
    digits = tuple(int(digit) for digit in str(abs(value)))
    exponent = -decimals
    return Decimal((sign, digits, exponent))


def hash_to_bool(passed_hash: str) -> bool:
    return hash_to_int(passed_hash) != 0


def hash_to_address(passed_hash: str) -> str:
    passed_hash = remove_0x_prefix(passed_hash)
    return f'0x{passed_hash[-40:]}'


def hash_to_words(passed_hash: str) -> List[str]:
    passed_hash = remove_0x_prefix(passed_hash)
    return [passed_hash[i:i + 64] for i in range(0, len(passed_hash), 64)]


def words_to_address(words: List[str]) -> str:
    return hash_to_address(words[0])


def words_to_int(words: List[str]) -> int:
    return hash_to_int(words[0])


def words_to_string(words: List[str]) -> str:
    string_length = hash_to_int(words[0]) * 2
    hex_string = ''.join(words[1:])
    return unhexlify(hex_string[:string_length]).decode()


def hash_to_list(
        passed_hash: str,
        serializer: Callable[[List[str]], Any] = words_to_address,
        dynamic_elements: bool = False
) -> List[Any]:
    return words_to_list(hash_to_words(passed_hash), serializer, dynamic_elements)


def words_to_list(
        words: List[str],
        serializer: Callable[[List[str]], Any],
        dynamic_elements: bool = False
) -> List[Any]:
    if not words:
        return []

    array_start = hash_to_int(words[0]) // 32
    array_length = hash_to_int(words[array_start])

    if array_length == 0:
        return []
    if dynamic_elements:
        element_positions = [
            hash_to_int(words[i]) // 32 for i in range(array_start + 1, array_start + 1 + array_length)
        ]
    else:
        element_positions = list(range(array_start - 1, array_start - 1 + array_length))

    result = []
    for i, position in enumerate(element_positions):
        start_position = array_start + 1 + position
        if i + 1 < len(element_positions):
            end_position = array_start + 1 + element_positions[i + 1]
            result.append(serializer(words[start_position:end_position]))
        else:
            result.append(serializer(words[start_position:]))
    return result


def get_signature_hash(signature: bytes) -> str:
    full_hash = keccak_256(signature).hexdigest()
    return hash_to_signature(full_hash)


def string_to_hash(passed_string: str) -> str:
    return represent_hash(bytes(passed_string, 'UTF-8'))



from typing import List, Union

from furl import furl

from defisdk.entities import AdapterBalance, ProtocolBalance, TokenBalance, TokenMetadata, AssetBalance
from defisdk.repositories.ethereum import BaseEthereumRepository
from defisdk.serializers import (
    defi_sdk_adapter_balances_to_entity,
    defi_sdk_protocol_balances_to_entity,
    defi_sdk_protocol_names_to_list_of_string,
    defi_sdk_token_adapter_names_to_list_of_string,
    defi_sdk_full_token_balance_to_entity
)
from defisdk.utils import get_signature_hash, hash_to_address, hash_to_list, remove_0x_prefix, string_to_hash


class DeFiSDKAPIRepository(BaseEthereumRepository):

    def __init__(self, ethereum_node_url: str, defisdk_registry: str):
        self._get_user_balance = get_signature_hash(b"getBalances(address)")
        self._get_adapter_balances = get_signature_hash(b"getAdapterBalances(address,address[])")
        self._get_protocol_balances = get_signature_hash(b"getProtocolBalances(address,string[])")
        self._get_protocol_names = get_signature_hash(b"getProtocolNames()")
        self._get_token_adapter_names = get_signature_hash(b"getTokenAdapterNames()")
        self._get_protocol_adapters = get_signature_hash(b"getProtocolAdapters(string)")
        self._get_token_adapter = get_signature_hash(b"getTokenAdapter(string)")
        self._get_full_token_balance = get_signature_hash(b"getFullTokenBalance(string,address)")
        self._get_final_full_token_balance = get_signature_hash(b"getFinalFullTokenBalance(string,address)")
        self._url = furl(ethereum_node_url)
        self._registry = defisdk_registry
        super().__init__()

    async def get_account_balance(
            self,
            address: str,
            block: Union[str, int] = 'latest'
    ) -> List[ProtocolBalance]:
        return await self._call(
            self._registry,
            self._get_user_balance + remove_0x_prefix(address).zfill(64),
            serializer=defi_sdk_protocol_balances_to_entity,
            block=block
        )

    async def get_adapter_balance(
            self,
            address: str,
            adapter: str,
            block: Union[str, int] = 'latest'
    ) -> AdapterBalance:
        words = [
            remove_0x_prefix(address).zfill(64),  # first argument: address
            remove_0x_prefix(hex(64)).zfill(64),  # position of the dynamic argument: [adapter]
            remove_0x_prefix(hex(1)).zfill(64),  # length of the dynamic argument: [adapter]
            remove_0x_prefix(adapter).zfill(64)  # second argument: [adapter]
        ]
        result = await self._call(
            self._registry,
            self._get_adapter_balances + ''.join(words),
            serializer=defi_sdk_adapter_balances_to_entity,
            block=block
        )
        return result[0]

    async def get_protocol_balance(
            self,
            address: str,
            protocol_name: str,
            block: Union[str, int] = 'latest'
    ) -> ProtocolBalance:
        words = [
            remove_0x_prefix(address).zfill(64),  # first argument: address
            remove_0x_prefix(hex(64)).zfill(64),  # position of the dynamic argument: [protocol[]]
            remove_0x_prefix(hex(1)).zfill(64),  # length of the dynamic argument: [protocol[]]
            remove_0x_prefix(hex(32)).zfill(64),  # position of the dynamic argument: [protocol]
            remove_0x_prefix(hex(len(protocol_name))).zfill(64),  # length of the dynamic argument: [protocol]
            remove_0x_prefix(string_to_hash(protocol_name)).ljust(64, '0')  # left-aligned argument
        ]
        result = await self._call(
            self._registry,
            self._get_protocol_balances + ''.join(words),
            serializer=defi_sdk_protocol_balances_to_entity,
            block=block
        )
        return result[0]

    async def get_protocol_names(
            self,
            block: Union[str, int] = 'latest'
    ) -> List[str]:
        return await self._call(
            self._registry,
            self._get_protocol_names,
            serializer=defi_sdk_protocol_names_to_list_of_string,
            block=block
        )

    async def get_token_adapter_names(
            self,
            block: Union[str, int] = 'latest'
    ) -> List[str]:
        return await self._call(
            self._registry,
            self._get_token_adapter_names,
            serializer=defi_sdk_token_adapter_names_to_list_of_string,
            block=block
        )

    async def get_protocol_adapter_names(
            self,
            protocol_name: str,
            block: Union[str, int] = 'latest'
    ) -> List[str]:
        words = [
            remove_0x_prefix(hex(32)).zfill(64),  # position of the dynamic argument
            remove_0x_prefix(hex(len(protocol_name))).zfill(64),  # length of the dynamic argument
            remove_0x_prefix(string_to_hash(protocol_name)).ljust(64, '0')  # left-aligned argument
        ]
        return await self._call(
            self._registry,
            self._get_protocol_adapters + ''.join(words),
            serializer=hash_to_list,
            block=block
        )

    async def get_token_adapter(
            self,
            token_adapter_name: str,
            block: Union[str, int] = 'latest'
    ) -> str:
        words = [
            remove_0x_prefix(hex(32)).zfill(64),  # position of the dynamic argument
            remove_0x_prefix(hex(len(token_adapter_name))).zfill(64),  # length of the dynamic argument
            remove_0x_prefix(string_to_hash(token_adapter_name)).ljust(64, '0')  # left-aligned argument
        ]
        return await self._call(
            self._registry,
            self._get_token_adapter + ''.join(words),
            serializer=hash_to_address,
            block=block
        )

    async def get_full_token_balance(
            self,
            token_type: str,
            token_address: str,
            block: Union[str, int] = 'latest'
    ) -> AssetBalance:
        words = [
            remove_0x_prefix(hex(64)).zfill(64),  # position of the dynamic argument
            remove_0x_prefix(token_address).zfill(64),  # second argument: address
            remove_0x_prefix(hex(len(token_type))).zfill(64),  # length of the dynamic argument
            remove_0x_prefix(string_to_hash(token_type)).ljust(64, '0')  # left-aligned argument
        ]
        return await self._call(
            self._registry,
            self._get_full_token_balance + ''.join(words),
            serializer=defi_sdk_full_token_balance_to_entity,
            block=block
        )

    async def get_token_components(
            self,
            token_type: str,
            token_address: str,
            block: Union[str, int] = 'latest'
    ) -> List[TokenBalance]:
        full_token_balance = await self.get_full_token_balance(token_type, token_address, block)
        return full_token_balance.underlying_token_balances

    async def get_token_metadata(
            self,
            token_type: str,
            token_address: str,
            block: Union[str, int] = 'latest'
    ) -> TokenMetadata:
        full_token_balance = await self.get_full_token_balance(token_type, token_address, block)
        return full_token_balance.base_token_balance.metadata



import itertools
from typing import Any, Dict, List, Optional

from defisdk.errors import EthereumNodeEmptyResponse, EthereumNodeNoResponse, EthereumNodeResponseException
from defisdk.repositories.base import BaseAPIRepository
from defisdk.utils import represent_address


class BaseEthereumRepository(BaseAPIRepository):

    def __init__(self, *args, **kwargs):
        self.request_counter = itertools.count()

    def prepare_rpc_request_body(self, method: str, params: Optional[List] = None) -> Dict[str, Any]:
        params = params or []
        return {"jsonrpc": "2.0", "method": method, "params": params or [], "id": next(self.request_counter)}

    async def call_method(self, method: str, params: Optional[List] = None) -> Optional[Dict[str, Any]]:
        headers = {'Content-Type': 'application/json'}
        data = self.prepare_rpc_request_body(method, params)
        node_answer = None

        try:
            node_answer = await self._post(url=self.url.url, headers=headers, data=data)
        except Exception as e:
            raise EthereumNodeResponseException(params, e.message, e.code, e.response)
        finally:
            if not node_answer.get('result'):
                raise EthereumNodeNoResponse(node_answer)
            if node_answer['result'] == '0x':
                raise EthereumNodeEmptyResponse()
            if node_answer and 'error' in node_answer:
                raise EthereumNodeResponseException(
                    params, node_answer['error']['message'], node_answer['error']['code'], None
                )

        return node_answer

    async def _call(self, _to: str, data: str, serializer=lambda x: x, block='latest'):
        block = hex(block) if isinstance(block, int) else block
        params = [{'to': represent_address(_to), 'data': data}, block]
        node_answer = await self.call_method('eth_call', params)
        return serializer(node_answer['result'])



#!/usr/bin/env python
from setuptools import convert_path, setup, find_packages

version = {}
version_path = convert_path('defisdk/version.py')
with open(version_path) as f:
    exec(f.read(), version)

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name='defisdk',
    version=version['__version__'],
    author='Alex Bash',
    author_email='alexey@zerion.io',
    description='DeFiSDK.py',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url='https://github.com/zeriontech/defi-sdk-py',
    packages=find_packages(),
    install_requires=[
        'aiohttp==3.6.2',
        'furl==2.1.0',
        'pysha3==1.0.2',
    ],
    python_requires='>=3.7',
)



class EthereumNodeResponseError(Exception):
    def __init__(self, message, code, body):
        super().__init__(message)
        self.message = message
        self.code = code
        self.body = body

    def __str__(self):
        return f'Error: {self.message}, {self.code} response: {self.body}'


class EthereumNodeResponseException(EthereumNodeResponseError):
    def __init__(self, params, message, code, data):
        super().__init__(message, code, data)
        self.params = params


class EthereumNodeNoResponse(EthereumNodeResponseError):
    def __init__(self, node_answer=None):
        super().__init__('No response', 500, node_answer)


class EthereumNodeEmptyResponse(EthereumNodeResponseError):
    def __init__(self):
        super().__init__('Empty response', 400, None)


class EthereumNodeInvalidResponse(EthereumNodeResponseError):
    def __init__(self, node_answer):
        super().__init__('Invalid response', 503, node_answer)



from dataclasses import dataclass
from decimal import Decimal
from typing import List


@dataclass(eq=True, frozen=True)
class TokenMetadata:
    address: str
    name: str
    symbol: str
    decimals: int


@dataclass(eq=True, frozen=True)
class TokenBalance:
    metadata: TokenMetadata
    balance: Decimal


@dataclass(eq=True, frozen=True)
class AssetBalance:
    base_token_balance: TokenBalance
    underlying_token_balances: List[TokenBalance]


@dataclass(eq=True, frozen=True)
class AdapterMetadata:
    address: str
    type: str


@dataclass(eq=True, frozen=True)
class AdapterBalance:
    metadata: AdapterMetadata
    asset_balances: List[AssetBalance]


@dataclass(eq=True, frozen=True)
class ProtocolMetadata:
    name: str
    description: str
    website_url: str
    icon_url: str
    version: int


@dataclass(eq=True, frozen=True)
class ProtocolBalance:
    metadata: ProtocolMetadata
    adapter_balances: List[AdapterBalance]
